/**
* Modified version of the fisheye package with support for the two implementations
* explored by gutwin.
* Type 0 : standard
*      1 : flat hemisphere
*      2 : Pyramid
*      3 : Only warp x
*/
export default function fisheye(distortion = 2, radius = 200,innerRadius=0,type=0) {
  const e = Math.exp(distortion)
  let k0 = e / (e - 1) * radius;
  let k1 = distortion / radius;
  const onlyForward = false

  return origin => item => {
    const dx = item.x - origin.x;
    const dy = item.y - origin.y;
    const distance = Math.sqrt(dx * dx + dy * dy);

    // This is the standard fisheye
    if(type===0){
      // too far away ? don't apply anything
      if(onlyForward && origin.x >item.x){
        return {
          x: item.x,
          y: item.y,
          scale: distance >= radius ? 1 : 10,
        };
      }
      if (distance === 0 || distance >= radius ) {
        return {
          x: item.x,
          y: item.y,
          scale: distance >= radius ? 1 : 10,
        };
      }

      const k = k0 * (1 - Math.exp(-distance * k1)) / distance * .75 + .25;
      return {
        x: origin.x + dx * k,
        y: origin.y + dy * k,
        scale: Math.min(k, 10),
      };
    }

    //Flat-hemisphere lens
    if(type===1){
      // too far away ? don't apply anything
      if (!distance || distance >= radius || distance <= innerRadius) {
        return {
          x: item.x,
          y: item.y,
          scale: 1,
        };
      }

      const k = k0 * (1 - Math.exp(-distance * k1)) / distance * .75 + .25;
      return {
        x: origin.x + dx * k,
        y: origin.y + dy * k,
        scale: Math.min(k, 10),
      };
    }

    //True Pyramid lens
    if(type===2){
      // too far away ? don't apply anything
      if (!distance || distance >= radius) {
        return {
          x: item.x,
          y: item.y,
          scale: 1,
        };
      }
      const k = k0 * (1 - Math.exp(-distance * k1)) / distance * .75 + .25;
      const d = dy/dx

      //Literally on the line
      if(item.y === origin.y){
        return {
          x: origin.x + dx * k,
          y: item.y,
          scale: Math.min(k, 10),
        };
      }
      //In top half
      else if(item.y < origin.y){
        //In first or fourth octant, warp x
        if(0>=d>=-1 || 0<=d<=1){
          return {
            x: origin.x + dx * k,
            y: item.y,
            scale: Math.min(k, 10),
          };
        }
        //In second or third octant, warp y
        else if(-1>=d || d>=1){
          return {
            x: item.x,
            y: origin.y + dy * k,
            scale: Math.min(k, 10),
          };
        }
      }
      //In lower half
      else if(item.y>origin.y){
        //In fifth or eigth octant, warp x
        if(0>=d>=-1 || 0<=d<=1){
          return {
            x: origin.x + dx * k,
            y: item.y,
            scale: Math.min(k, 10),
          };
        }
        //In sixth or seventh octant, warp y
        else if(d<=-1 || d>=1){
          return {
            x: item.x,
            y: origin.y + dy * k,
            scale: Math.min(k, 10),
          };
        }
      }

    }

    //only distort x
    if(type===3){
      // too far away ? don't apply anything
      if (!distance || distance >= radius) {
        return {
          x: item.x,
          y: item.y,
          scale: distance >= radius ? 1 : 10,
        };
      }

      const k = k0 * (1 - Math.exp(-distance * k1)) / distance * .75 + .25;
      return {
        x: origin.x + dx * k,
        y: item.y,
        scale: Math.min(k, 10),
      };
    }
  }
}
