
import branches from '../../videos/branches'
import {state, findBranchByID} from './state'
import {BranchPoint, BranchSegment, Point } from './shapes/'
import {clearButtons} from './branchButtons'

// Init playback stuff
const canvas = document.getElementById('canvas')
const video = document.getElementById('video-player')
const overlayedButtons = document.getElementById('overlayedButtons')
const canvasButtons = document.getElementById('canvasButtons')
const ctx = canvas.getContext('2d')
const heightPercent = 0.09                                              // The height percentage of the playback bar
initSize()                                                              // Sets the size of elements that depend on the playback bra height
const middle = canvas.height / 2                                        // The middle of the playback bar
const backgroundOpacity = 0.3                                           // Opacity of the playback bar background
export const pixelsPerSecond = ctx.canvas.width/state.longestPath       // The length of on second in pixels. Set by the canvas width and longest path
const branchWidth = ctx.canvas.height/11                                // 11 Is arbitrarily set after some testing
const textSize = ctx.canvas.height/5                                    // 5 is arbitrarily set after some testing
var mouseInCanvas = false                                               // Controls when to hide the playback bar
const f = state.branchFunction                                          // The function used for the branches

// Lists of everything we want to draw
const segments = {}                   // All drawable segments
const branchPoints = []               // All drawable branchpoints
createShapes()                        // Create all the drawables
let fisheyeCenter = new Point(100000, 1000000)   // Init fisheye center
let canvasScale = 1                   // The current scale applied to the canvas, used for mouse pos
let canvasYPan = 0                    // The current y translation applied to the canvas, used for mouse pos
let canvasXPan = 0                    // The current x translation applied to the canvas, used for mouse pos
initEvents()                          // Initialise all mouse events

/**
 * Updates the playback bar
 */
export function updatePlaybackBar () {

  const {onlyDrawUpcoming} = state
  if (resetCanvas()===0) return
  doTransformations()
  //Drawing branches and branch points
  if(onlyDrawUpcoming){
      drawUpcoming('branches')
      drawUpcoming('branchpoints')
  }
  else {
      drawAll('branches')
      drawAll('branchpoints')
  }
}

/**
 * Draws all of the given mode
 * @param mode branches or branchpoints
 */
function drawAll(mode){
  branches.forEach(branch => {
    const branchID = branch.id
    if(mode==="branches")drawBranch(branchID)
    else if(mode==="branchpoints"){
        if(branch.branches.length === 0) return
        drawBranchPoint(branchID)
    }
  })
}

/**
 * Draws only upcoming of the given mode.
 * This method is recursive
 * @param mode branches or branchpoints
 * @param branchID This should not be supplied unless the root id isn't A
 * @param i This should not be applied as this controls the drawdepth when the current branch is reached
 */
function drawUpcoming(mode,branchID='A',i=-99999){
    const {upcomingBranches, currentBranch, drawDepth, parentLookupTable} = state
    const branch = findBranchByID(branchID)
    if(i<=drawDepth && ( // Draw if allowed
        branchID === currentBranch ||
        upcomingBranches.includes(branchID) ||
        upcomingBranches.some(r=> parentLookupTable[branchID].includes(r)) ||
        segments[branchID].playTime>0)
    )
    {
      if(currentBranch === branchID){ //If we reached the current branch, start counting levels drawn.
          i=0
      }
      if(mode === "branches"){
          drawBranch(branchID)
      }
      else if(mode === "branchpoints"){
          if(branch.branches.length === 0) return
          drawBranchPoint(branchID)
      }

      if(branch.branches.length === 0) return
        branch.branches.forEach(child=>{
          drawUpcoming(mode,child,i+1)
      })
    }
    return
}

/**
 * Draws the branch with the given id. Called from the drawAll or drawUpcoming.
 * @param branchID the id of the branch to be drawn
 */
function drawBranch(branchID){
    if(!branchID) return
    const {fisheye, follow } = state
    updateBranch(branchID)
    if (fisheye) {
        segments[branchID].drawFisheySegmentedOfFunction(ctx, fisheyeCenter)
    } else if(follow) {
        segments[branchID].drawFisheySegmentedOfFunction(ctx, new Point(100000,100000))
    }
    else{
        segments[branchID].drawFisheySegmentedOfFunction(ctx, new Point(100000,100000))
    }
}

/**
 * Draws the branchpoint with the given id. Called from the drawAll or drawUpcoming.
 * @param branchID the id of the branchpoint to be drawn
 */
function drawBranchPoint(branchID){
    const {fisheye} = state
    if (fisheye) {
        branchPoints[branchID].drawFisheye(ctx, fisheyeCenter)
    } else {
        branchPoints[branchID].draw(ctx)
    }
}


/**
 * Updates the playtime and buffer of the given branch
 *
 * @param branchID id of the branch you want to update
 */
function updateBranch(branchID) {
  const {playTime,currBuff,isHighlight,drawLetters,highlightOnHover,matchWithLetters} = state
  segments[branchID].playTime = playTime[branchID]
  segments[branchID].buffer =  currBuff[branchID]
  if(highlightOnHover)segments[branchID].highlight = isHighlight[branchID]
  if(matchWithLetters)segments[branchID].drawLetter = drawLetters[branchID]
}

/**
 * Creates branches and branchpoints. This is recursive so don't touch it.
 * @param branchID
 * @param x
 * @param y
 * @param endY
 * @param curveFunction
 * @param yOffset
 * @param textOffset
 */
function createShapes (branchID = 'A', x = 0, y = middle, endY = middle, curveFunction = f,yOffset=middle/2,textOffset = 0) {
  const branch = findBranchByID(branchID)
  const endX = x + branch.length * pixelsPerSecond

  const branchSegment = new BranchSegment(x, y, endX, endY, branch.length, curveFunction, branchID, branchWidth,textSize,textOffset)
  segments[branchID] = branchSegment

  const children = branch.branches
  //const yOffset = Math.abs(y)-Math.abs(endY)

  // Has children => call this function again
  if (children.length>0) {
    branchPoints[branchID] = (new BranchPoint(endX, endY, branchID,6))

    if (children.length === 2) {
      createShapes(children[0], endX, endY, endY - yOffset,f,yOffset/2,0)
      createShapes(children[1], endX, endY, endY + yOffset,f,yOffset/2,1)
    } else if (children.length === 3) {
      createShapes(children[0], endX, endY, endY - yOffset,f,yOffset/2,0)
      createShapes(children[1], endX, endY, endY,f,0,1)
      createShapes(children[2], endX, endY, endY + yOffset,f,yOffset/2,2)
    }
  }

}

/**
 * Resets the canvas returns 0 if the mouse isn't in the player.
 * This is so the background isn't drawn in this case.
 * @returns {number}
 */
function resetCanvas() {
  ctx.resetTransform()
  ctx.clearRect(-100, -100, 2000, 1000)
  ctx.globalAlpha = backgroundOpacity
  ctx.fillStyle = '#fff'
  if(!mouseInCanvas) return 0
  ctx.fillRect(-100, -100, 2000, 1000)
  ctx.globalAlpha = 1
}

/**
 * Does calculations for the transformations and applies them
 */
function doTransformations() {
  const {follow, fisheye, currentBranch, mouseEye} = state
  const currentSegment = segments[currentBranch]
  const {x} = currentSegment.playedEnd.get()

  const {fisheyeOffset} = state
  //Calculate fisheyeposition
  const fisheyeX = -x + fisheyeOffset
  const fisheyeY = currentSegment.getYfromX(x+fisheyeOffset)
  const chaseX = x +fisheyeOffset
  const chaseY = fisheyeY
  if(mouseEye){
    document.getElementById('playerContainer').onmousemove  = function(e) {
      fisheyeCenter = new Point(getMousePos(canvas,e).x,getMousePos(canvas,e).y)
    }
  }else if(fisheye){
    fisheyeCenter = new Point(chaseX-fisheyeOffset, currentSegment.startPoint.y +chaseY)
  }
  // Transforms
  if (fisheye && follow) {
    const scale = 1
    ctx.scale(scale, scale)
    ctx.translate(fisheyeX, -currentSegment.startPoint.y + middle/scale - fisheyeY)
    canvasScale = scale
    canvasXPan = fisheyeX
    canvasYPan = -currentSegment.startPoint.y + middle/scale -fisheyeY

  }
  else if (follow) {
    const scale = 1.5
    ctx.scale(scale, scale)
    ctx.translate(fisheyeX,  - currentSegment.startPoint.y + middle/scale - fisheyeY)
    canvasScale = scale
    canvasXPan = fisheyeX
    canvasYPan = -currentSegment.startPoint.y + middle/scale -fisheyeY
  }
  else {
    ctx.scale(1, 1)
  }
}

function getMousePosOld(canvas, evt) {
  var rect = canvas.getBoundingClientRect();
  return {
      x: (evt.clientX - rect.left) / (rect.right - rect.left) * canvas.width,
      y: (evt.clientY - rect.top) / (rect.bottom - rect.top) * canvas.height
  };
}

/**
 * gets mouse position relative to the canvas. Takes scale and pan into account
 * @param canvas the canvas
 * @param evt the event
 * @returns {{x: number, y: number}} relative x and y coordinates
 */
function getMousePos(canvas, evt) {
  return {
      x: ((evt.clientX -canvas.offsetLeft))/canvasScale -canvasXPan,
      y: ((evt.clientY - canvas.offsetTop-3))/canvasScale -canvasYPan
  };
}

/**
 * Initialises mouse events
 */
function initEvents(){
  //Makes the bar disappear when the
  //mouse is outside the player
  overlayedButtons.onmouseenter =
    canvasButtons.onmouseenter =
      video.onmouseenter =
        canvas.onmouseenter = function () {
          mouseInCanvas = true
        }
  overlayedButtons.onmouseleave =
    canvasButtons.onmouseleave =
      video.onmouseleave =
        canvas.onmouseleave = function () {
          mouseInCanvas = false
        }
  const {highlightOnBranchHover,clickableBranches} = state

  if(highlightOnBranchHover) {
    canvas.onmousemove = canvasButtons.onmousemove = function (e) {
      const {isHighlight, upcomingBranches,nextBranch} = state
      if(nextBranch)return
      let {x: posX, y: posY} = getMousePos(canvas,e)
      upcomingBranches.forEach(branch => {
        if (segments[branch].advancedIncludes(posX, posY,fisheyeCenter) && upcomingBranches.includes(branch)) {
          isHighlight[branch] = true
        }
        else {
          isHighlight[branch] = false
        }
      })
    }
  }
  if(clickableBranches){
    canvas.onmousedown = canvasButtons.onmousedown = function (e) {
      let {upcomingBranches,nextBranch} = state
      if(nextBranch)return
      let {x: posX, y: posY} = getMousePos(canvas,e)
      upcomingBranches.forEach(branch => {
        if (segments[branch].advancedIncludes(posX, posY,fisheyeCenter) && upcomingBranches.includes(branch)) {
          state.nextBranch = branch
          clearButtons()
        }
        else {
          state.isHighlight[branch] = false
        }
      })
    }
  }
}

/**
 * Initialises the size of objects affected by the bar height.
 */
function initSize(){
  ctx.canvas.width = video.clientWidth
  ctx.canvas.height = video.clientHeight*heightPercent
  canvas.style.bottom = video.clientHeight*(1+heightPercent)+5 +"px" //this 5 is fucking mystery
  overlayedButtons.style.height = (1-heightPercent)*100+"%"
  canvasButtons.style.height = heightPercent*100+"%"
}