import {Point} from './index'
const branchBackgroundColor = '#111'
const branchBufferColor = '#858585'
const branchPlayedColor = '#CC1C13'
const branchHighlightColor = '#02bfff'
const innerSegmentScale = 0.7
const baseTextOffset = 8 //This results in the text being one fith of the branch length from the start point
const fidelity = 1 //With this set to 1 you draw 1 segment per second. With 2 you draw 2 and so on.
// The "buffer-pipe"
/** A branch segment representation that is drawable
 *
 */
export class BranchSegment {
  constructor (startX, startY, endX, endY, bufferMax = 32 , curveFunction, id, height=8, textSize, textOffset=0, buffer = 0, playTime = 0) {
    this.startPoint = new Point(startX, startY)
    this.endPoint = new Point(endX, endY)
    //this.scale = 1
    this.playedEnd = this.startPoint
    this.height = height
    this.playTime = playTime
    this.buffer = buffer
    this.bufferMax = bufferMax
    this.curveFunction = curveFunction
    this.id = id
    this.highlight = false

    //For letters
    this.drawLetter = false
    this.textSize = textSize
    this.textOffset = textOffset
    let textX = startX + (endX-startX)/baseTextOffset +this.textSize*this.textOffset
    let textY = endY  + this.height/2
    this.textPoint = new Point(textX,textY)
  }

    /**
     * Draws the segment with regard to a fisheye centre point as well as the segements curvefunction
     * @param ctx the canvas
     * @param fisheyePoint the centre point of the fisheye lens
     * @param height of the segment, if you want to override it
     */
  drawFisheySegmentedOfFunction (ctx, fisheyePoint,height = this.height) {
    this.playedEnd = getEndpoint(this.startPoint, this.endPoint, this.playTime / this.bufferMax)

    const width = (this.endPoint.x - this.startPoint.x) / (this.bufferMax*fidelity)
    for (let i = 0; i < this.bufferMax*fidelity; i++) {

        var x = i * width
        var y = this.getYfromX(x)
        const segmentStart = this.startPoint.add(x, y)

        x += width
        y = this.getYfromX(x)
        const segmentEnd = this.startPoint.add(x, y)
        // draw background
        //const highlight = true
        if(this.highlight){
          drawFisheyeLine(ctx, segmentStart, segmentEnd, fisheyePoint, height, branchHighlightColor)
        }else{
          drawFisheyeLine(ctx, segmentStart, segmentEnd, fisheyePoint, height, branchBackgroundColor)
        }


        // draw buffer
        if (i < this.buffer * fidelity) {
            drawFisheyeLine(ctx, segmentStart, segmentEnd, fisheyePoint, height * innerSegmentScale, branchBufferColor)
        }

        //draw playtime
        if (i < this.playTime * fidelity) {
            let increment = 1

            if (this.playTime * fidelity - 1 < i && i < this.playTime * fidelity) {
                increment = this.playTime % 1
            }
            var x = (increment + i) * width
            var y = this.getYfromX(x)
            const bufferEnd = this.startPoint.add(x, y)

            drawFisheyeLine(ctx, segmentStart, bufferEnd, fisheyePoint, height * innerSegmentScale, branchPlayedColor)
        }
    }
    if(this.drawLetter){
      drawFisheyeLetter(ctx,this.textPoint,fisheyePoint,this.textSize,this.id)
    }

  }

  /**
   * calculates the y value for a given x through
   * scaling the function as described in our thesis
   *
   * @param x the x value
   * @returns {number} the y value of x
   */
  getYfromX(x){
      return ((this.endPoint.y-this.startPoint.y)
          /this.curveFunction(
              this.endPoint.x - this.startPoint.x
          ))*this.curveFunction(x)
  }

  /**
   * Calculates if
   * @param posX
   * @param posY
   * @returns {boolean}
   */
  includes(posX,posY){
    if(posX>this.startPoint.x && posX<this.endPoint.x){
      if(posY<this.endPoint.y+this.height/2 && posY> this.endPoint.y-this.height/2){
        return true
      }else return false
    }else return false
  }

  /**
   * Advanced version of the includes function that actually
   * takes the curve function into consideration.
   * @param posX
   * @param posY
   * @param fisheyePoint
   * @returns {boolean}
   */
  advancedIncludes(posX,posY,fisheyePoint){
    if(posX>this.startPoint.x && posX<this.endPoint.x){
      const width = (this.endPoint.x - this.startPoint.x) / (this.bufferMax*fidelity)
      for (let i = 0; i < this.bufferMax*fidelity; i++) {
        let x = i * width
        let y = this.getYfromX(x)
        const boxStart = this.startPoint.add(x, y)

        x += width
        y = this.getYfromX(x)
        const boxEnd = this.startPoint.add(x, y)

        //Utan fisheye
        /*
        const {x: startX, y: startY} = boxStart
        const {x: endX, y: endY} = boxEnd
        const currHeight = this.height
        */

        //Med fisheye
        const {x: startX, y: startY, scale: startScale} = boxStart.fisheye(fisheyePoint)
        const {x: endX, y: endY, scale: endScale} = boxEnd.fisheye(fisheyePoint)
        const averageScale = Math.min((startScale + endScale) / 2, 1.5)
        const currHeight = this.height*averageScale;

        let vector = new Point(endX-startX,endY-startY)
        let alpha = (Math.PI/2 - Math.atan(vector.x/Math.abs(vector.y)))
        if(vector.y >0) alpha = -alpha
        const {u: startU} = doRotationTranslation(startX,startY,alpha)
        const {u: endU, v: endV } = doRotationTranslation(endX,endY,alpha)
        const {u: clickU, v: clickV } = doRotationTranslation(posX,posY,alpha)

        if(clickU<endU && clickU>startU) {
          if (clickV < endV + currHeight/1.2 && clickV > endV - currHeight/1.2) {
            return true
          }
        }
      }
    }else return false
  }

}

/**
 * Translates a given x and y coordinate in respect to the given angle
 * @param x
 * @param y
 * @param a the angle (counter clockwise
 * @returns {{u: number, v: number}} The u and v coordinates
 */
function doRotationTranslation(x,y,a){
  return {
    u: x*Math.cos(a) - y*Math.sin(a),
    v: x*Math.sin(a) + y*Math.cos(a)
  }
}

/**
 * This is used for calculating the end position for the playtime bar
 * Should not be used for anything else
 * @param startPoint
 * @param endPoint
 * @param procent
 * @returns {Point}
 */

function getEndpoint (startPoint, endPoint, procent) {
  const yDiff = endPoint.y - startPoint.y
  const xDiff = endPoint.x - startPoint.x

  procent = Math.min(procent, 1)

  const newEndX = startPoint.x + xDiff * procent
  const newEndY = startPoint.y + yDiff * procent

  return new Point(newEndX, newEndY)
}

/**
 * Draws the given segment with respect to the fisheye
 * @param ctx the canvas
 * @param startPoint of the line to be drawn
 * @param endPoint of the line to be drawn
 * @param fisheyePoint, the centerpoint of the fisheye lens
 * @param scale of the bar
 * @param lineWidth of the bar
 * @param color of the bar
 */
function drawFisheyeLine (ctx, startPoint, endPoint, fisheyePoint, lineWidth = 1, color = '#111') {
  const {x: startX, y: startY, scale: startScale} = startPoint.fisheye(fisheyePoint)
  const {x: endX, y: endY, scale: endScale} = endPoint.fisheye(fisheyePoint)
  const averageScale = Math.min((startScale + endScale) / 2, 1.5)
  // Set new styles
  ctx.lineWidth = lineWidth * averageScale
  ctx.strokeStyle = color

  ctx.beginPath()
  ctx.moveTo(startX, startY)
  ctx.lineTo(endX, endY)
  ctx.stroke()
}
function drawFisheyeLetter(ctx,point,fisheyePoint,textSize,id){
  const {x: posX, y: posY, scale: scale} = point.fisheye(fisheyePoint)
  ctx.fillStyle = "white"
  ctx.textAlign = "center"
  ctx.strokeStyle = "black"
  ctx.font = (scale * textSize) + "px Comic Sans MS"
  ctx.strokeText(id, posX, posY)
  ctx.fillText(id, posX, posY)
}
