import player from './branchedPlayer'
import {buffer} from '../SegmentCache'
import branches from '../../videos/branches'

// Global state for branched player
export const state = {
  currentBranch: 'A',                     // The current branch
  upcomingBranches: ['B', 'C'],           // The upcoming branches
  nextBranch: null,                       // The selected upcomming branch-
  time: 0,                                // Current position in time
  timeForBranch: 32,                      // TODO förklara denna
  segment: 1,                             // Current segment, not used?
  playTime: {},                           // Current play time for branches
  currBuff: {},                           // Current buffer for branches
  isHighlight: {},                        // Which branches to highlight
  drawLetters: {},                        // Which branches to draw letters for
  fisheye: false,                         // enable fisheye effect
  fisheyeType: 0,                         // Fisheye type, see gutwinsFisheye package
  follow: false,                           // follow the current position in the playback bar
  mouseEye: false,                        // Centre the fisheye on the mouse position
  highlightOnHover: true,                 // Highlight the matching branch off the button you are hovering over
  highlightOnBranchHover: true,          // Highlight the branch you are hovering in the playback bar if it is upcoming
  clickableBranches: true,               // Enables you to click
  useOverlayButtons: true,                // Use buttons placed on top of the video
  useCanvasButtons: false,                // Use buttons placed on top of the playback bar
  matchWithLetters: true,                 // Match the buttons and branches by drawing text on the
  onlyDrawUpcoming: false,                // Only draw upcoming branches
  drawDepth: 1,                           // The draw depth of upcoming branches
  fisheyeOffset: 120,                     // The fisheye offset from the current position in the playback bar
  branchFunction: function (x){           // The function that the branches follow
    if(x!=0)                              //
      return Math.atan(x)                 //
    else return 0                         //
  },                                      //
  parentLookupTable: createLookupTable(), // Lookup table to find the parent of a given branch
  longestPath: findLongestPath()          // The longest path available in seconds
}

/**
 * Prints the current state of playback
 */
export function printState () {
  console.log(
    'Playback state : [' , '\n',
    'currentBranch : ', state.currentBranch, '\n',
    'upcomingBranches : ', state.upcomingBranches, '\n',
    'nextBranch : ', state.nextBranch, '\n',
    'time : ', state.time, '\n',
    'timeForBranch : ', state.timeForBranch, '\n',
    'segment : ', state.segment, '\n',
    'playTime :' , state.playTime, '\n',
    ']'
  )
}

/**
 * Updates all the state variables that need updating
 */
export function updateState () {
  const currentTime = player.time()
  const currentBranch = findBranchFromTime(currentTime)

  // Update state object
  state.time = currentTime
  state.currentBranch = currentBranch.id
  state.upcomingBranches = currentBranch.branches ? currentBranch.branches : []
  state.timeForBranch = currentBranch.startTime + currentBranch.length - currentTime
  state.segment = findSegmentFromTime(currentTime)
  state.playTime = calculatePlayTimeFromTime(currentTime)
  state.currBuff = buffer
}

/**
 * Finds a branch from a given time
 * @param time the time you want to find a branch for
 * @returns {*}
 */
export function findBranchFromTime (time) {
  const branch = branches.find(branch => {
    if (time >= branch.startTime && time <= branch.startTime + branch.length) {
      return true
    }
  })

  return branch
}

/**
 * Returns the current segment from a given time. Assumes 4 second segments
 * @param time the time you want to find the segment for
 * @returns {number} the segment
 */
export function findSegmentFromTime (time) {
  return Math.floor(time / 4) + 1
}

/**
 * Finds the branch for the given id
 * @param branchID for the branch you want to find
 * @returns {*} the branch
 */
export function findBranchByID (branchID) {
  return branches.find(branch => branch.id === branchID)
}

/**
 * Generates a parent lookup table.
 */
function createLookupTable() {
  let temp = {}
  for(let i=0;i<branches.length;i++){
    let branchID = branches[i].id
    temp[branchID] = []
    branches.forEach(branch =>{
      if(branch.branches.includes(branchID)){
        temp[branchID].push(branch.id)
        if(temp[branch.id].length>0){
          temp[branchID].push.apply(temp[branchID],temp[branch.id])
        }
      }
    })
  }
  return temp
}

/**
 * Recursive function that finds the longest possible path through a video
 * @param time do not pass a value for this unless you want to start at a later point than 0 seconds
 * @param branchID do not pass a value for this unless your root isn't called A.
 * @returns {number} the longest path in seconds.
 */
function findLongestPath(time=0,branchID = 'A') {
  var branch = findBranchByID(branchID)
  time += branch.length
  var temp = time
  if(branch.branches.length>0) {
    branch.branches.forEach(childID => {
      var ret = findLongestPath(temp, childID)
      if (ret > time) time = ret
    })
  }
  return time
}
/**
 * Voodoo magic to set the playtime for all branches
 * @param time
 */
function calculatePlayTimeFromTime (time) {
  const currentID = findBranchFromTime(time).id
  const parents = state.parentLookupTable[currentID]
  const playTime = {}

  branches.forEach(branch => {
    // If current branch then we have played at part of it
    if (branch.id === currentID) {
      playTime[branch.id] = time - branch.startTime
    }
    // If its our parent branch we have already played all of it
    else if (parents.includes(branch.id)) {
      playTime[branch.id] = branch.length
    }
    // Otherwise we haven't played anything of that branch
    else {
      playTime[branch.id] = 0
    }
  })

  return playTime
}
